/***************************************************************************

  CTreeView.cpp

  The TreeView class

  (c) 2000-2001 Benot Minisini <gambas@freesurf.fr>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/


#define __CTREEVIEW_CPP

#include "common.h"

#include <qapplication.h>
#include <qheader.h>
#include <qpalette.h>
#include <qscrollview.h>
#include <qlistview.h>
#include <qobjcoll.h>

#include "gambas.h"

#include "CWidget.h"
#include "CPicture.h"
#include "CTreeView.h"

static void *CLASS_TreeViewItem;

static long EVENT_Select;    /* selection change */
static long EVENT_Click;     /* simple click */
static long EVENT_Activate;  /* double click */
static long EVENT_Expand;
static long EVENT_Collapse;


/***************************************************************************

  class MyListViewItem

***************************************************************************/

void MyListViewItem::initData(void)
{
  item.picture = NULL;
  item.key = NULL;
  item.item = this;
  item.ob.klass = (long)CLASS_TreeViewItem;
  item.ob.ref = 1;
}


MyListViewItem::MyListViewItem(QListView *parent)
: QListViewItem(parent)
{
  initData();
}

MyListViewItem::MyListViewItem(QListView *parent, MyListViewItem *after)
: QListViewItem(parent, after)
{
  initData();
}

MyListViewItem::MyListViewItem(MyListViewItem *parent)
: QListViewItem(parent)
{
  initData();
}

MyListViewItem::MyListViewItem(MyListViewItem *parent, MyListViewItem *after)
: QListViewItem(parent, after)
{
  initData();
}

MyListViewItem::~MyListViewItem()
{
  GB.Unref((void **)&item.picture);
  GB.FreeString(item.key);
}

void MyListViewItem::setPicture(CPICTURE *pict)
{
  GB.StoreObject(pict, (void **)&item.picture);
  setPixmap(0, *pict->pixmap);
}


/***************************************************************************

  TreeView

***************************************************************************/

BEGIN_METHOD_VOID(CTREEVIEW_init)

  CLASS_TreeViewItem = GB.FindClass("Node");

END_METHOD


BEGIN_METHOD(CTREEVIEW_new, void *parent)

  QListView *wid = new QListView(CONTAINER(PARAM(parent)));

  QObject::connect(wid, SIGNAL(selectionChanged()), &CTreeView::manager, SLOT(selected()));
  QObject::connect(wid, SIGNAL(doubleClicked(QListViewItem *)), &CTreeView::manager, SLOT(activated()));
  QObject::connect(wid, SIGNAL(clicked(QListViewItem *)), &CTreeView::manager, SLOT(clicked(QListViewItem *)));
  QObject::connect(wid, SIGNAL(expanded(QListViewItem *)), &CTreeView::manager, SLOT(expanded(QListViewItem *)));
  QObject::connect(wid, SIGNAL(collapsed(QListViewItem *)), &CTreeView::manager, SLOT(collapsed(QListViewItem *)));

  //OBJECT(CTREEVIEW)->lock = 0;

  CWIDGET_new(wid, (void *)_object, "TreeView");

  OBJECT(CTREEVIEW)->dict = new QAsciiDict<MyListViewItem>;
  OBJECT(CWIDGET)->background = QColorGroup::Base;

  wid->addColumn("");
  wid->header()->hide();
  wid->show();

END_METHOD


BEGIN_METHOD_VOID(CTREEVIEW_free)

  delete OBJECT(CTREEVIEW)->dict;

  CALL_METHOD_VOID(CWIDGET_delete);

END_METHOD


BEGIN_METHOD_VOID(CTREEVIEW_clear)

  OBJECT(CTREEVIEW)->dict->clear();
  QLISTVIEW(_object)->clear();

END_METHOD


BEGIN_METHOD(CTREEVIEW_add, GB_STRING key; GB_STRING text; CPICTURE *picture; GB_STRING parent; GB_STRING after)

  MyListViewItem *item;
  QListView *wid = QLISTVIEW(_object);
  char *key = GB.ToZeroString(PARAM(key));
  MyListViewItem *parent = NULL;
  MyListViewItem *after = NULL;
  char *parent_key;

  /*
  if (GB.IsMissing(4))
    qDebug("CTREEVIEW_add: key = %s  parent = NULL", key);
  else
    qDebug("CTREEVIEW_add: key = %s  parent = %s", key, GB.ToZeroString(PARAM(parent)));
  */

  if (key == NULL)
  {
    GB.Error("Bad key");
    return;
  }

  item = (*OBJECT(CTREEVIEW)->dict)[key];
  if (item != NULL)
  {
    GB.Error("Key already used");
    return;
  }

  if (!GB.IsMissing(4))
  {
    parent_key = GB.ToZeroString(PARAM(parent));
    parent = NULL;
    if (parent_key)
      parent = (*OBJECT(CTREEVIEW)->dict)[parent_key];
    if (parent == NULL)
    {
      GB.Error("Parent item does not exist");
      return;
    }
  }

  if (!GB.IsMissing(5))
  {
    after = (*OBJECT(CTREEVIEW)->dict)[GB.ToZeroString(PARAM(parent))];
    if (after == NULL)
    {
      GB.Error("After item does not exist");
      return;
    }
  }

  if (parent == NULL)
  {
    if (after == NULL)
      item = new MyListViewItem(wid);
    else
      item = new MyListViewItem(wid, after);
  }
  else
  {
    if (after == NULL)
      item = new MyListViewItem(parent);
    else
      item = new MyListViewItem(parent, after);
  }

  OBJECT(CTREEVIEW)->dict->insert(key, item);

  item->setText(0, GB.ToZeroString(PARAM(text)));

  item->item.key = GB.CopyString(key, PARAM(key).len);

  if (!GB.IsMissing(3))
    item->setPicture(PARAM(picture));

  GB.ReturnObject(&item->item);

END_METHOD


BEGIN_METHOD(CTREEVIEW_remove, GB_STRING key)

  MyListViewItem *item;
  char *key = GB.ToZeroString(PARAM(key));
  //QListView *wid = QLISTVIEW(_object);

  item = CTreeView::getItem(OBJECT(CTREEVIEW), key);
  if (!item)
    return;

  if (OBJECT(CTREEVIEW)->widget.ob.ref > 1)
  {
    GB.Error("Item is referenced");
    return;
  }

  OBJECT(CTREEVIEW)->dict->remove(key);
  delete item;

END_METHOD


BEGIN_METHOD(CTREEVIEW_get, GB_STRING key)

  MyListViewItem *item;
  char *key = GB.ToZeroString(PARAM(key));
  //QListView *wid = QLISTVIEW(_object);

  item = CTreeView::getItem(OBJECT(CTREEVIEW), key);
  if (!item)
    return;

  GB.ReturnObject(&item->item);

END_METHOD


BEGIN_PROPERTY(CTREEVIEW_mode)

  if (READ_PROPERTY)
    GB.ReturnInteger(QLISTVIEW(_object)->selectionMode());
  else
    QLISTVIEW(_object)->setSelectionMode((QListView::SelectionMode)PROPERTY(long));

END_PROPERTY


BEGIN_PROPERTY(CTREEVIEW_current)

  MyListViewItem *item = (MyListViewItem *)(QLISTVIEW(_object)->currentItem());

  if (item == 0)
    GB.ReturnNull();
  else
    GB.ReturnObject(&item->item);

END_PROPERTY


BEGIN_PROPERTY(CTREEVIEW_background)

  QWidget *wid = QWIDGET(_object);
  QPalette pal(wid->palette());
  QColorGroup::ColorRole role;

  role = QColorGroup::Base;

  if (READ_PROPERTY)
    GB.ReturnInteger(pal.color(QPalette::Active, QColorGroup::Base).rgb() & 0xFFFFFF);
  else
  {
    pal.setColor(role, QColor((QRgb)PROPERTY(long)));
    wid->setPalette(pal);
  }

END_PROPERTY



/*
BEGIN_METHOD(CLISTVIEW_remove, long pos)

  QLISTVIEW(_object)->removeItem(PARAM(pos));

END_METHOD

BEGIN_METHOD_VOID(CLISTVIEW_sort)

  QLISTVIEW(_object)->sort();

END_METHOD


BEGIN_PROPERTY(CLISTVIEW_index)

  if (READ_PROPERTY)
    GB.ReturnInteger(QLISTVIEW(_object)->currentItem());
  else
    QLISTVIEW(_object)->setCurrentItem(PROPERTY(long));

END_PROPERTY


BEGIN_PROPERTY(CLISTVIEW_current)

  int current = QLISTVIEW(_object)->currentItem();

  if (current < 0)
    GB.ReturnString(NULL);
  else
    GB.ReturnString(QLISTVIEW(_object)->text(current));

END_PROPERTY


*/


BEGIN_PROPERTY(CTREEVIEWITEM_key)

  GB.ReturnZeroString(OBJECT(CTREEVIEWITEM)->key);

END_PROPERTY


BEGIN_PROPERTY(CTREEVIEWITEM_picture)

  if (READ_PROPERTY)
    GB.ReturnObject(OBJECT(CTREEVIEWITEM)->picture);
  else
    OBJECT(CTREEVIEWITEM)->item->setPicture(PROPERTY(CPICTURE *));

END_PROPERTY


BEGIN_PROPERTY(CTREEVIEWITEM_selected)

  MyListViewItem *it = OBJECT(CTREEVIEWITEM)->item;

  if (READ_PROPERTY)
    GB.ReturnBoolean(it->isSelected());
  else
    it->listView()->setSelected(it, PROPERTY(char) != 0);

END_PROPERTY


BEGIN_PROPERTY(CTREEVIEWITEM_expanded)

  MyListViewItem *it = OBJECT(CTREEVIEWITEM)->item;

  if (READ_PROPERTY)
    GB.ReturnBoolean(it->isOpen());
  else
    it->setOpen(PROPERTY(char) != 0);

END_PROPERTY


BEGIN_PROPERTY(CTREEVIEWITEM_text)

  MyListViewItem *it = OBJECT(CTREEVIEWITEM)->item;

  if (READ_PROPERTY)
    GB.ReturnZeroString(it->text(0));
  else
    it->setText(0, PROPERTY(char *));

END_PROPERTY


BEGIN_PROPERTY(CTREEVIEWITEM_parent)

  MyListViewItem *lvi = (MyListViewItem *)(OBJECT(CTREEVIEWITEM)->item->parent());

  if (lvi == 0)
    GB.ReturnNull();
  else
    GB.ReturnObject(&lvi->item);

END_PROPERTY


BEGIN_PROPERTY(CTREEVIEWITEM_child)

  MyListViewItem *lvi = (MyListViewItem *)(OBJECT(CTREEVIEWITEM)->item->firstChild());

  if (lvi == 0)
    GB.ReturnNull();
  else
    GB.ReturnObject(&lvi->item);

END_PROPERTY


BEGIN_PROPERTY(CTREEVIEWITEM_next)

  MyListViewItem *lvi = (MyListViewItem *)(OBJECT(CTREEVIEWITEM)->item->nextSibling());

  if (lvi == 0)
    GB.ReturnNull();
  else
    GB.ReturnObject(&lvi->item);

END_PROPERTY


BEGIN_METHOD_VOID(CTREEVIEWITEM_ensure_visible)

  const MyListViewItem* item = OBJECT(CTREEVIEWITEM)->item;

  item->listView()->ensureItemVisible(item);

END_METHOD



/***************************************************************************

  Descriptions

***************************************************************************/

GB_DESC CTreeViewItemDesc[] =
{
  GB_DECLARE("Node", 0), GB_HOOK_NEW(NULL), GB_HOOK_FREE(NULL),

  GB_PROPERTY_READ("Key", "s", CTREEVIEWITEM_key),
  GB_PROPERTY("Picture", "Picture", CTREEVIEWITEM_picture),
  GB_PROPERTY("Selected", "b", CTREEVIEWITEM_selected),
  GB_PROPERTY("Text", "s", CTREEVIEWITEM_text),
  GB_PROPERTY_READ("Parent", "Node", CTREEVIEWITEM_parent),
  GB_PROPERTY_READ("Next", "Node", CTREEVIEWITEM_next),
  GB_PROPERTY_READ("Child", "Node", CTREEVIEWITEM_child),
  GB_PROPERTY("Expanded", "b", CTREEVIEWITEM_expanded),
  GB_METHOD("EnsureVisible", NULL, CTREEVIEWITEM_ensure_visible, NULL),

  GB_END_DECLARE
};


GB_DESC CTreeViewDesc[] =
{
  GB_DECLARE("TreeView", sizeof(CTREEVIEW)), GB_INHERITS("Control"),

  GB_CONSTANT("None", "i", QListView::NoSelection),
  GB_CONSTANT("Single", "i", QListView::Single),
  GB_CONSTANT("Multi", "i", QListView::Multi),
  GB_CONSTANT("Extended", "i", QListView::Extended),

  GB_STATIC_METHOD("_init", NULL, CTREEVIEW_init, NULL),
  GB_METHOD("_new", NULL, CTREEVIEW_new, "Container;"),
  GB_METHOD("_free", NULL, CTREEVIEW_free, NULL),

  GB_PROPERTY("Mode", "i", CTREEVIEW_mode),

  //GB_PROPERTY("Columns", "i", CLISTVIEW_columns),

  GB_METHOD("_get", "Node", CTREEVIEW_get, "(Key)s"),

  GB_METHOD("Clear", NULL, CTREEVIEW_clear, NULL),
  GB_METHOD("Add", "Node", CTREEVIEW_add, "(Key)s(Text)s[(Picture)Picture;(Parent)s(After)s]"),
  GB_METHOD("Remove", NULL, CTREEVIEW_remove, "(Key)s"),
  //GB_METHOD("Sort", NULL, CLISTVIEW_sort, NULL),

  /*
  GB_METHOD("Find", "i", CLISTVIEW_find, 1, 1, "s"),
  GB_METHOD("Select", NULL, CLISTVIEW_select, 1, 1, "i"),
  GB_METHOD("Invert", NULL, CLISTVIEW_invert, 0, 0, NULL),
  */
  //GB_PROPERTY_READ("Text", ".ListViewText", CLISTVIEW_text_picture),
  //GB_PROPERTY_READ("Picture", ".ListViewPicture", CLISTVIEW_text_picture),

  GB_PROPERTY_READ("Current", "Node", CTREEVIEW_current),

  GB_CONSTANT("_Properties", "s", CTREEVIEW_PROPERTIES),
  GB_CONSTANT("_DefaultEvent", "s", "Click"),

  GB_EVENT("Select", NULL, NULL, &EVENT_Select),
  GB_EVENT("Activate", NULL, NULL, &EVENT_Activate),
  GB_EVENT("Click", NULL, NULL, &EVENT_Click),
  GB_EVENT("Expand", NULL, "(Item)Node;", &EVENT_Expand),
  GB_EVENT("Collapse", NULL, "(Item)Node;", &EVENT_Collapse),

  GB_END_DECLARE
};


/***************************************************************************

  class CTreeView

***************************************************************************/

CTreeView CTreeView::manager;


MyListViewItem *CTreeView::getItem(CTREEVIEW *treeview, char *key)
{
  MyListViewItem *item = (*treeview->dict)[key];

  if (item == 0)
  {
    GB.Error("Unknown item '%s'", key);
    return NULL;
  }

  return item;
}

/*
void CListView::removeItem(CLISTVIEW *listview, char *key)
{
  listview->dict.remove(key);
}
*/

void CTreeView::destroy(void)
{
}

void CTreeView::selected(void)
{
  RAISE_EVENT(EVENT_Select, NULL);
}

void CTreeView::activated(void)
{
  RAISE_EVENT(EVENT_Activate, NULL);
}

void CTreeView::clicked(QListViewItem *it)
{
  if (it != 0)
  {
    //CTREEVIEWITEM *item = &(((MyListViewItem *)it)->item);
    RAISE_EVENT(EVENT_Click, NULL);
  }
}

void CTreeView::expanded(QListViewItem *it)
{
  CTREEVIEWITEM *item = &(((MyListViewItem *)it)->item);

  RAISE_EVENT(EVENT_Expand, (void *)&item);
}

void CTreeView::collapsed(QListViewItem *it)
{
  CTREEVIEWITEM *item = &(((MyListViewItem *)it)->item);

  RAISE_EVENT(EVENT_Collapse, (void *)&item);
}
